@@ -15,7 +15,11 @@ import typing
 import tempfile
 from collections import namedtuple
 import yaml
+import warnings
+import urllib.parse
+import shutil
 
+from .engine_intf import GenericStreamWrapper, BytesIOWrapper
 from .proto_handlers import load_protocol_handler
 from .utils.connect import connect_to_database, dataframe_to_sql
 from .utils.dbconn import sql_enquote_id
@@ -29,6 +33,11 @@ try:
 except ImportError:
     mlflow = None
 
+try:
+    import dask.dataframe
+except ImportError:
+    dask = None
+
 from .metadata import Metadata
 
 # this environment variable says how DSLibrary should be implemented
@@ -768,7 +777,7 @@ class DSLibrary(object):
             f_w.write(chunk)
         return "-"
 
-    def load_dataframe(self, resource_name: str, **kwargs):
+    def load_dataframe(self, resource_name: (str, io.RawIOBase, typing.Iterable), **kwargs):
         """
         Load a pandas dataframe from various types of input resources.
 
@@ -781,73 +790,122 @@ class DSLibrary(object):
             sql_table - name of table to fully load
             nosql_collection - name of a NoSQL collection to fully load
             run_data - True to read using open_run_data()
+            dask - True to load into a dask dataframe, or an integer to load into dask if file size is >= to threshold.
 
         :param resource_name:   The name of an input resource, or of a local file, or the URI of an external resource.
         """
-        # check for specifically supplied data
-        supplied = (self._spec.get("data") or {}).get(resource_name)
-        if supplied is not None and hasattr(supplied, "columns"):
-            return supplied
-        # map requested resource (resource_name) through input spec and override defaults (kwargs)
-        uri, open_args = self._xlt_resource(resource_name, **kwargs)
+        # look up pre-supplied data
+        # do translation of named resources
+        if isinstance(resource_name, str):
+            # check for specifically supplied data
+            supplied = (self._spec.get("data") or {}).get(resource_name)
+            if supplied is not None and hasattr(supplied, "columns"):
+                return supplied
+            # map requested resource (resource_name) through input spec and override defaults (kwargs)
+            uri, open_args = self._xlt_resource(resource_name, **kwargs)
+        else:
+            uri = getattr(resource_name, "name") if hasattr(resource_name, "name") else ""
+            open_args = kwargs
         # arguments that are specific to load_dataframe() and should not go through to open_resource()
         fallback_to_text = open_args.pop("fallback_to_text", False)
-        use_dask = open_args.pop("dask", None)
+        use_dask = open_args.pop("dask", None) or 0
+        if use_dask and not dask:
+            warnings.warn("Dask was requested but is not installed", ImportWarning)
+            use_dask = False
         format_options = open_args.pop("format_options", None) or {}
         format = open_args.pop("format", None)
         if format:
             format_options = dict(format_options)
             format_options["format"] = format
-        # detect SQL
-        if open_args.get("sql") or open_args.get("sql_table"):
-            sql = open_args.pop("sql", None)
-            table = open_args.pop("sql_table", None)
-            if table:
-                sql = f"SELECT * from {sql_enquote_id(table)}"
-            db_conn = self.get_sql_connection(resource_name, **open_args)
-            try:
-                c = db_conn.cursor()
-                c.execute(sql)
-                col_names = [col[0] for col in c.description]
-                df = pandas.DataFrame(data=c, columns=col_names)
-            finally:
-                db_conn.close()
-            return df
-        # detect NoSQL
-        if open_args.get("nosql_collection"):
-            collection = open_args.pop("nosql_collection")
-            db_conn = self.get_nosql_connection(resource_name, **open_args)
-            rows = list(db_conn.query(collection))
-            return pandas.DataFrame(rows)
-        # TODO we should offer a validation option that checks the data types against the schema
-        # format can be specified, or file extension will be used
-        file_extension = find_url_extension(uri)
-        if not file_extension:
-            self._default_format_for_aliases(resource_name, open_args)
+        if "#" in uri:
+            # URI can contain formatting arguments
+            uri, uri_args = uri.split("#", maxsplit=1)
+            for k, v in urllib.parse.parse_qs(uri_args).items():
+                format_options[k] = v[0] if len(v) == 1 else v
+        if isinstance(resource_name, str):
+            # detect SQL
+            if open_args.get("sql") or open_args.get("sql_table"):
+                sql = open_args.pop("sql", None)
+                table = open_args.pop("sql_table", None)
+                if table:
+                    sql = f"SELECT * from {sql_enquote_id(table)}"
+                db_conn = self.get_sql_connection(resource_name, **open_args)
+                try:
+                    c = db_conn.cursor()
+                    c.execute(sql)
+                    col_names = [col[0] for col in c.description]
+                    df = pandas.DataFrame(data=c, columns=col_names)
+                    if use_dask:
+                        mem_size = df.memory_usage()
+                        if mem_size > use_dask:
+                            partition_size = max(use_dask, 40000000)
+                            df = dask.dataframe.from_pandas(df, npartitions=min(mem_size//partition_size, 2))
+                finally:
+                    db_conn.close()
+                return df
+            # detect NoSQL
+            if open_args.get("nosql_collection"):
+                collection = open_args.pop("nosql_collection")
+                db_conn = self.get_nosql_connection(resource_name, **open_args)
+                rows = list(db_conn.query(collection))
+                return pandas.DataFrame(rows)
+            # TODO we should offer a validation option that checks the data types against the schema
+            # format can be specified, or file extension will be used
+            file_extension = find_url_extension(uri)
+            if not file_extension:
+                self._default_format_for_aliases(resource_name, open_args)
+            # this method will be used to open the stream...
+            use_opener = self.open_run_data if open_args.pop("run_data", None) else self.open_resource
+        elif hasattr(resource_name, "read"):
+            stream = resource_name
+            orig_mode = "r" if isinstance(stream, io.TextIOBase) else "rb"
+            count = [0]
+            def open_existing_stream(filename: str, mode: str='rb', **kwargs):
+                if count[0]:
+                    if hasattr(stream, "seek") and hasattr(stream, "seekable") and stream.seekable():
+                        stream.seek(0)
+                    else:
+                        raise Exception("internal usage error, stream is not seekable")
+                count[0] += 1
+                s_out = GenericStreamWrapper(stream, close=False)
+                if mode != orig_mode:
+                    if mode == 'r':
+                        s_out = io.TextIOWrapper(s_out, 'utf-8')
+                    else:
+                        s_out = BytesIOWrapper(s_out)
+                return s_out
+            use_opener = open_existing_stream
+        else:
+            # list, dict, etc.
+            # FIXME what to do with all the other arguments?  dask, at least could be applied
+            return pandas.DataFrame(resource_name)
         # sample-taker
-        use_opener = self.open_run_data if open_args.pop("run_data", None) else self.open_resource
         def sampler():
             sample_size = open_args.pop("sample_size", None) or 30000
             if sample_size < 0:
                 return
             with use_opener(uri, mode='rb', **open_args) as f_r:
                 return f_r.read(sample_size)
-        '''
-        # TODO we could measure the size, but this doesn't add much to our format sniffing and could cause problems on
-        #   some data sources
         def sizer():
-            with self.open_resource(resource_name, mode='rb', **kwargs) as f_r:
+            with use_opener(resource_name, mode='rb') as f_r:
                 if not hasattr(f_r, "seek") or not f_r.seekable():
                     return
                 f_r.seek(0, 2)
                 return f_r.tell()
-        '''
+        # FIXME detect non-seekable stream and inhibit sampling/sizing
         # sniff format
+        use_sizer = use_dask > 1
         format_info = detect_format(
-            uri, sampler=sampler, sizer=None, supplied_options=format_options, fallback_to_text=fallback_to_text
+            uri, sampler=sampler, sizer=sizer if use_sizer else None, supplied_options=format_options, fallback_to_text=fallback_to_text
         )
         format = format_info.format
         read_args = format_info.read_args
+        # choose dask based on threshold
+        if use_dask > 1:
+            if format_info.size is not None:
+                use_dask = format_info.size > use_dask
+            else:
+                use_dask = 0
         # various formats and their read function
         if use_dask:
             use_dask = format in DASK_READ_FORMATS
@@ -871,11 +929,28 @@ class DSLibrary(object):
                 if format == "xlsx" and not f_r.seekable():
                     # the XLSX reader requires its stream to be seekable so we download to a temporary file
                     with tempfile.NamedTemporaryFile(mode='wb') as f_tmp:
-                        f_tmp.write(f_r.read())
+                        # chunked copy
+                        shutil.copyfileobj(f_r, f_tmp)
                         f_tmp.flush()
                         with open(f_tmp.name, mode='rb') as f_r2:
                             return read_method(f_r2, **read_args)
-                return read_method(f_r, **read_args)
+                # the 'hdf' implementation doesn't accept a stream, only a filename
+                if format == "hdf" and hasattr(f_r, "name") and ":" not in f_r.name:
+                    return read_method(f_r.name, **read_args)
+                if format == "hdf":
+                    with tempfile.NamedTemporaryFile(mode='wb') as f_tmp:
+                        shutil.copyfileobj(f_r, f_tmp)
+                        f_tmp.flush()
+                        return read_method(f_tmp.name, **read_args)
+                try:
+                    return read_method(f_r, **read_args)
+                except UnicodeDecodeError as err:
+                    raise DSLibraryDataFormatException(f"Invalid character encoding: {err.reason}: encoding={err.encoding}, offset={err.start} - set 'encoding' or 'encoding_errors' options to compensate", offset=err.start)
+                except ValueError as err:
+                    if format == "json":
+                        raise DSLibraryDataFormatException(f"JSON format error: {str(err)}")
+                    else:
+                        raise DSLibraryDataFormatException(f"{str(err)}")
         # TODO support read from folder, i.e. folder of CSVs, parquet, etc.
         raise DSLibraryException(f"file type not supported: {format}")
 
@@ -934,6 +1009,16 @@ class DSLibraryException(Exception):
     """
 
 
+class DSLibraryDataFormatException(DSLibraryException):
+    """
+    Problems with data
+    """
+    def __init__(self, message: str, offset: int=None):
+        super(DSLibraryDataFormatException, self).__init__(message, offset)
+        self.message = message
+        self.offset = offset
+
+
 class DSLibraryCommunicationError(DSLibraryException):
     """
     Communication errors.
