@@ -116,7 +116,7 @@ class DSLibraryViaMMLibrary(DSLibrary):
             return super(DSLibraryViaMMLibrary, self).log_metric(metric_name, metric_value, step)
         self._mm.save_kpi(metric_name, metric_value)
 
-    def open_run_data(self, filename: str, mode: str='rb') -> io.IOBase:
+    def open_run_data(self, filename: str, mode: str='rb', **kwargs) -> io.IOBase:
         """
         The method provided by mmlibrary to save 'run data' only stores a single chunk of binary data.  We
         store JSON in that data, and translate calls to read and write files such that they read and write
@@ -165,6 +165,12 @@ class DSLibraryViaMMLibrary(DSLibrary):
             more["_expected_system_type"] = system_type
         if "_read_only" in sig.parameters:
             more["_read_only"] = not for_write
+        db_or_bucket = kwargs.get("database") or kwargs.get("bucket_name")
+        if db_or_bucket:
+            if "_db_or_bucket" in sig.parameters:
+                more["_db_or_bucket"] = db_or_bucket
+            else:
+                raise DSLibraryException("Incompatible use of 'database' or 'bucket_name' argument -- override is not supported by this version of mmlibrary")
         return method(resource_name, **more)
 
     def set_evaluation_result(self, success: bool, **kwargs) -> None:
