@@ -9,7 +9,7 @@ import tempfile
 import yaml
 
 from dslibrary import DSLibrary
-from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS, DSLibraryException
+from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS, DSLibraryException, DSLibraryDataFormatException
 from dslibrary.metadata import Metadata
 from dslibrary.transport.to_local import DSLibraryLocal
 
@@ -536,3 +536,93 @@ class TestFront(unittest.TestCase):
             ('CREATE TABLE IF NOT EXISTS tbl1 (col1 INTEGER)', []),
             ('INSERT INTO tbl1 (col1) VALUES (%s), (%s), (%s)', [1, 2, 3])
         ])
+
+    def test_load_dataframe__dask_autodetect(self):
+        """
+        Size-based threshold.
+        """
+        with tempfile.NamedTemporaryFile(suffix=".csv") as f_tmp:
+            f_tmp.write(b"x\n1\n2")
+            f_tmp.flush()
+            class MyInst(DSLibrary):
+                def _opener(self, path: str, mode: str, **kwargs):
+                    return io.BytesIO(b"x\n1\n2")
+            # specifically don't use dask
+            df = MyInst().load_dataframe(f_tmp.name, dask=False)
+            self.assertIsInstance(df, pandas.DataFrame)
+            # specifically do use dask
+            df = MyInst().load_dataframe(f_tmp.name, dask=True)
+            assert hasattr(df, "dask")
+            # high threshold, don't use dask
+            df = MyInst().load_dataframe(f_tmp.name, dask=1000)
+            self.assertIsInstance(df, pandas.DataFrame)
+            # low threshold, do use dask
+            df = MyInst().load_dataframe(f_tmp.name, dask=3)
+            assert hasattr(df, "dask")
+
+    def test_load_dataframe__from_stream(self):
+        """
+        An already opened stream can be supplied instead of a named entity.
+        """
+        stream = io.BytesIO(b"x\n1\n2\n12")
+        df = DSLibrary().load_dataframe(stream)
+        self.assertEqual(list(df.x), [1, 2, 12])
+        stream = io.StringIO("x\n1\n2\n12")
+        df = DSLibrary().load_dataframe(stream)
+        self.assertEqual(list(df.x), [1, 2, 12])
+
+    def test_load_dataframe__from_iterable(self):
+        """
+        Data can be supplied to turn into a dataframe.
+        """
+        df = DSLibrary().load_dataframe({"x": [11, 12, 13]})
+        self.assertEqual(list(df.x), [11, 12, 13])
+        df = DSLibrary().load_dataframe([{"x": 9}, {"x": 8}])
+        self.assertEqual(list(df.x), [9, 8])
+
+    def test_hash_in_uri_for_formmatting_argumets(self):
+        """
+        Since the '#' has no effect on information returned from a URL, it is a safe place to place formatting arguments.
+        """
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                assert path == "x", "the '#' is removed before opening the file"
+                return io.BytesIO(b"a,b\n1,2\n2,3")
+        # here the default format sniffer guesses there are two columns
+        df = MyInst().load_dataframe("x")
+        self.assertEqual(list(df.columns), ["a", "b"])
+        # here we force the delimited to be a linefeed and now there is only one column
+        df = MyInst().load_dataframe("x#delimiter=%0a")
+        self.assertEqual(list(df.columns), ["a,b"])
+
+    def test_hdf(self):
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                return open(path, mode)
+        fn = os.path.join(os.path.dirname(__file__), "test_data/my.hdf")
+        df = MyInst().load_dataframe(fn)
+        assert list(df.columns) == ["x"]
+        assert list(df.x) == [1, 2, 3]
+
+    def test_report_encoding_error(self):
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                return io.BytesIO(b'a\n1\n2\n3\n\xC0\xC0\xFF\n4\n5\n6')
+        try:
+            MyInst().load_dataframe("x.csv")
+            self.fail("expected error")
+        except DSLibraryDataFormatException as err:
+            assert "invalid start byte" in err.message
+            assert "offset=8" in err.message
+        df = MyInst().load_dataframe("x.csv", format_options={"encoding": 'iso8859-1'})
+        self.assertEqual(list(df.columns), ["a"])
+
+    def test_report_json_format_error(self):
+        class MyInst(DSLibrary):
+            def _opener(self, path: str, mode: str, **kwargs):
+                return io.BytesIO(b'{"x":' + b' '*1000)
+        try:
+            MyInst().load_dataframe("x.json")
+            self.fail("expected error")
+        except DSLibraryDataFormatException as err:
+            assert "JSON format error" in err.message
